def recommend_model(dataset, target):
    """Recommend which models to use for training"""
    THRESHOLD = 0.01

    if isinstance(dataset["target"], str):
        return {"recommended_model": "classification"}

    else:
        num_unique_values = len(dataset[target].unique())
        num_rows = len(dataset[target])

        if num_unique_values / num_rows < THRESHOLD:
            return {"recommended_model": "classification"}
        else:
            return {"recommended_model": "regression"}



from collections import defaultdict


class Graph:
    def __init__(self, vertices):
        self.graph = defaultdict(list)
        self.V = vertices

    def addEdge(self, u, v):
        self.graph[u].append(v)

    def topologicalSortUtil(self, v, visited, stack):
        visited[v] = True
        for i in self.graph[v]:
            if visited[i] == False:
                self.topologicalSortUtil(i, visited, stack)
        stack.append(v)

    def topologicalSort(self):
        visited = [False] * self.V
        stack = []
        for i in range(self.V):
            if visited[i] == False:
                self.topologicalSortUtil(i, visited, stack)
        return stack[::-1]

    def isCyclicUtil(self, v, visited, recStack):
        visited[v] = True
        recStack[v] = True
        for neighbour in self.graph[v]:
            if visited[neighbour] == False:
                if self.isCyclicUtil(neighbour, visited, recStack) == True:
                    return True
            elif recStack[neighbour] == True:
                return True
        recStack[v] = False
        return False
 
    def isCyclic(self):
        visited = [False] * (self.V + 1)
        recStack = [False] * (self.V + 1)
        for node in range(self.V):
            if visited[node] == False:
                if self.isCyclicUtil(node,visited,recStack) == True:
                    return True
        return False



import pandas as pd
import boto3
from ..util.data_types import get_dataframe_types
import boto3


class Dataset:
    def __init__(self, url) -> None:
        self.file_name = url
        self.dataframe = self.__load_dataset(url)

    def __load_dataset(self, url: str) -> pd.DataFrame:
        print("Loading dataset from S3: ", url)
        df = pd.read_csv("https://const-bucket.s3.ap-south-1.amazonaws.com/" + url)
        df.columns = df.columns.str.replace(' ', '_')
        return df

    def parse_dataset(self):
        data_dict = get_dataframe_types(self.dataframe)
        data_dict["rows"] = len(self.dataframe.axes[0])
        return data_dict



from .experiment import Experiment
import string
import random


class Deployment:
    def generate_code(self, experiment_id: str):
        experiment = Experiment(experiment_id=experiment_id)
        with open("deployment.py", "w") as f:
            f.write("import requests")
            f.write("\nimport os")
            f.write("\nimport pandas as pd")
            f.write("\nfrom pydantic import BaseModel, Field")
            f.write("\nfrom huble import Experiment,Dataset")
            f.write("\nfrom typing import Union")
            f.write("\nimport joblib")
            f.write('\nexperiment_id = os.getenv("EXPERIMENT_ID")')
            f.write(f"\nexperiment = Experiment(experiment_id)")
            f.write("\n")
            f.write("\nclass RequestBody(BaseModel):")
            for column in experiment.input_format["columns"]:
                if column["name"] != "index":
                    f.write(
                        f"\n\t{''.join(random.choices(string.ascii_lowercase,k=7))}: Union[int, float,str] = Field(None,alias=\"{column['name']}\")"
                    )
            f.write("\n")
            f.write(f"\nmodel_url = experiment.modelURL")
            f.write('\nmodel_content = requests.get(f"https://const-bucket.s3.ap-south-1.amazonaws.com/{model_url}")')
            f.write("\nopen('{model_url}.joblib', 'wb').write(model_content.content)")
            f.write("\nmodel = joblib.load('{model_url}.joblib')")
            f.write("\n")
            f.write("\n")
            f.write("\ndef predict(X):")
            f.write("\n\treturn model.predict(X)")



import evalml
import json
import numpy as np
import pandas as pd
from evalml.data_checks import DefaultDataChecks
from evalml.automl import AutoMLSearch
from evalml.pipelines.utils import generate_pipeline_code
from ..util.data_types import get_dataframe_types


def convert_to_json(my_dict):
    for key, value in my_dict.items():
        if isinstance(value, dict):
            convert_to_json(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    convert_to_json(item)
                elif isinstance(item, (np.integer, np.floating)):
                    item = item.item()
        elif isinstance(value, (np.integer, np.floating)):
            my_dict[key] = value.item()
    return json.dumps(my_dict)


def automl(data, target_column, task):
    # split into X and y
    X = data.drop(target_column, axis=1)
    y = data[target_column]

    data_dict = get_dataframe_types(X)

    #Predict task type
    if task=="classification":
      n = len(pd.unique(data[target_column]))
    
      if n==2:
        task="binary"
      else:
        task="multiclass"
     

    #train test split
    X_train, X_test, y_train, y_test = evalml.preprocessing.split_data(X, y, problem_type=task)

    # Datachecks
    objective = ""
    if task == "binary":
        objective = objective + "log loss binary"
    elif task == "multiclass":
        objective = objective + "log loss multiclass"
    elif task == "regression":
        objective = objective + "R2"
    data_checks = DefaultDataChecks(task, objective)
    messages = data_checks.validate(X_train, y_train)

    errors = [message for message in messages if message["level"] == "error"]
    warnings = [message for message in messages if message["level"] == "warning"]
    data_issues = []
    for warning in warnings:
        data_issues.append("Warning:" + warning["message"])

    for error in errors:
        data_issues.append("Error:" + error["message"])



    #AutoMl Search
    automl = AutoMLSearch(X_train=X_train, y_train=y_train, problem_type=task)
    automl.search()

    # Pipeline Rankings
    rankings = automl.rankings
    rankings_json = rankings.to_json(orient="table")

    ids=[]
    for i in range(6):
        ids.append([rankings['id'].iloc[i], rankings['ranking_score'].iloc[i]])    
    
    pipelines=[]
 
    for i, rank in ids:
        pip={}
        pipeline = automl.get_pipeline(i)
        pip["id"]=i
        pip["ranking_score"]=rank
        pip['name']=pipeline.name
        pip['pipeline']=str(pipeline)
        
        pipeline.fit(X_train, y_train) 
        
        #metrics
        objectives=[]
        if task =='binary':
            objectives.extend(['mcc binary',
                                'log loss binary',
                                'gini',
                                'auc',
                                'recall',
                                'precision',
                                'f1',
                                'balanced accuracy binary',
                                'accuracy binary'])
        elif task == 'multiclass':
            objectives.extend(['mcc multiclass',
                                'log loss multiclass',
                                'auc weighted',
                                'auc macro',
                                'auc micro',
                                'recall weighted',
                                'recall macro',
                                'recall micro',
                                'precision weighted',
                                'precision macro',
                                'precision micro',
                                'f1 weighted',
                                'f1 macro',
                                'f1 micro',
                                'balanced accuracy multiclass',
                                'accuracy multiclass'])
        elif task =='regression':
            objectives.extend(['expvariance',
                                'maxerror',
                                'medianae',
                                'mse',
                                'mae',
                                'r2',
                                'mean squared log error',
                                'root mean squared log error',
                                'root mean squared error',])
        pip['metrics'] = dict(pipeline.score(X_test, y_test, objectives)) 
        
        #Feature importance for best pipeline
        feature_imp = pipeline.feature_importance
        feature_imp_json = feature_imp.to_json(orient="table")
        pip['feature_importance']=feature_imp_json

        # generating code for best_pipeline
        pip["code"] = generate_pipeline_code(pipeline)

        pipelines.append(pip)

    automl_result={}
    automl_result['Data Issues']=data_issues
    automl_result['Data Issue Messages']=messages
    automl_result['Task']=task
    automl_result['Pipelines']=pipelines
    automl_result['input_format']=data_dict

    automl_json = convert_to_json(automl_result)
    automl_json = automl_json.replace("\\", "")
    automl_json = automl_json.replace('"{"schema', '{"schema')
    automl_json = automl_json.replace('", "code', ' ,"code')

    return automl_json



from ..sklearn.process.handler import PreprocessHandler
from ..sklearn.graph import Graph
from .typed_params import get_typed_params

def preprocess_code_generation(graph, colab=False):
    g = Graph(len(graph["nodes"]))
    graph_dict = {}
    for i in graph["nodes"]:
        graph_dict[(i["data"]["name"])] = i
    map = {}
    j = 0
    while j < (len(graph["nodes"])):
        for i in graph["nodes"]:
            map[(i["id"])] = j
            j += 1
    map2 = {y: x for x, y in map.items()}
    for i in graph["edges"]:
        g.addEdge(map[i["source"]], map[i["target"]])
    res = g.topologicalSort()
    steps = []
    for i in res:
        steps.append(map2[i])
    steps_list = {}
    for i in steps:
        for j in range(len(graph["nodes"])):
            if i == graph["nodes"][j]["id"]:
                steps_list[(graph["nodes"][j]["data"]["name"])] = graph["nodes"][j][
                    "data"
                ]["node_type"]
    if colab:
        output_file = "/content/preprocess_code.py"
    else:
        output_file = "preprocess.py"
    preprocess_handler = PreprocessHandler()
    with open(output_file, "w") as f:
        f.write("import huble\n")
        f.write("def preprocess_code(data):\n")
        for i in steps_list:
            node = graph_dict[i]
            params = get_typed_params(node["data"]["parameters"])
            if steps_list[i] == "preprocess":
                f.write(
                    "\t"
                    + preprocess_handler.return_function(
                        function_name=node["data"]["name"], params=params
                    )
                )
                f.write("\n")

        f.write("\n\treturn data")



import pandas as pd
from sklearn.graph import Graph
from collections import Counter


def verify_pipeline(graph):
    issues = []
    steps = {}
    for node in graph["nodes"]:
        steps[node["data"]["name"]] = node["data"]["node_type"]
        if node["data"]["node_type"] == "primary_dataset" or node["data"]["node_type"] == "model" or node["data"]["node_type"] == "classification_model" or node["data"]["node_type"] == "regression_model":
            if "target" not in node:
                issues.append(node["data"]["name"] + " Node not connected")
        elif node["data"]["node_type"] == "evaluate_model":
            counter =0
            for i in graph['edges']:
              if i["target"]=="998":
                  counter+=1
            if counter<2:
                issues.append(node["data"]["name"] + " Node not connected")
        else:
            id = node['id']
            
            inputs = len(node['data']['inputParameters']['types'])
            outputs = len(node['data']['outputParameters']['types'])
            input=0
            output=0
            for i in graph['edges']:
              if i['target']==id:
                input+=1
              elif i['source']==id:
                output+=1
            
            if input!=inputs or output!=outputs:
                issues.append(node["data"]["name"] + " Node not connected")

    value_counter = Counter(steps.values())
    key_counter = Counter(steps.keys())
  
    if 'Train Model' not in steps:
      issues.append('Train Model Node Missing')
    elif key_counter['Train Model']>1:
      issues.append('More than one Train Model Node found')
    if 'Train-Test Split' not in steps:
      issues.append('Train-Test Split Node Missing')
    elif key_counter['Train-Test Split']>1:
      issues.append('More than one Train-Test Split Node found')
    if 'Evaluate Model' not in steps:
      issues.append('Evaluate Model Node Missing')
    elif key_counter['Evaluate Model']>1:
      issues.append('More than one Evaluate Model Node found')
    if 'model' not in steps.values() and 'classification_model' not in steps.values() and 'regression_model' not in steps.values():
      issues.append('Model Node Missing')
    elif value_counter['model']>1:
      issues.append('More than one Model Node found')
    if 'primary_dataset' not in steps.values():
      issues.append('Dataset Node Missing')
    elif value_counter['primary_dataset']>1:
      issues.append('More than one Dataset Node found')

    g = Graph(len(graph["nodes"]))
    graph_dict = {}
    for i in graph["nodes"]:
        graph_dict[(i["data"]["name"])] = i
    map = {}
    j = 0
    while j < (len(graph["nodes"])):
        for i in graph["nodes"]:
            map[(i["id"])] = j
            j += 1
  
    for i in graph["edges"]:
        g.addEdge(map[i["source"]], map[i["target"]])

    if g.isCyclic() == 1:
        issues.append("Graph contains cycle")
    else:
        pass

    return issues



from huble.sklearn.metrics import log_metrics
from sklearn import inspection
import numpy as np
import pandas as pd

def evaluate_model(model, training_dataset, test_dataset, target_column, task_type):
    X = training_dataset.drop([target_column], axis=1)
    y = training_dataset[target_column]
    if task_type == "clustering":        
        score = model.score(X)
        importances = np.repeat(score, X.shape[1]) 
        labels = model.labels_
        metrics = log_metrics(y_true=None, y_pred=None, task=task_type, X=training_dataset, labels=labels)       
    else:
        y_test = test_dataset[target_column]
        X_test = test_dataset.drop([target_column], axis=1)
        y_pred = model.predict(X_test)
        metrics = log_metrics(y_true=y_test, y_pred=y_pred, task=task_type, X=None, labels=None)
        result = inspection.permutation_importance(model, X, y, n_repeats=10, random_state=42)
        importances = result.importances_mean

    feature_importance = pd.DataFrame({'Feature': X.columns, 'Importance': importances})
    feature_importance = feature_importance.sort_values('Importance', ascending=False).reset_index(drop=True)
    
    
    return metrics, feature_importance.to_json()




def inference_pip(graph):
    id = "0"
    prediction_id = ""
    for node in graph["nodes"]:
        if node["data"]["node_type"] == "primary_dataset":
            node["data"]["name"] = "Web API"
        elif node["data"]["name"] == "Train-Test Split":
            prediction_id = node["id"]
            node["data"]["name"] = "Predictions"
            node["data"]["value"] = "Predictions"
            node["data"]["inputParameters"]["types"].append("Trained Model")
            node.pop("target")
            node["data"].pop("outputParameters")
            node["data"]["color"] = "green.400"
        elif node["data"]["name"] == "Train Model":
            id = node["id"]
        elif node["data"]["node_type"] == "essential":
            graph["nodes"].remove(node)

    for node in graph["nodes"]:
        if (
            node["data"]["node_type"] == "classification_model"
            or node["data"]["node_type"] == "regression_model"
            or node["data"]["node_type"] == "clusterring_model"
        ):
            # get node id and make an edge to predictions wala node
            node["data"]["name"] = "Trained " + node["data"]["name"]
            node["data"]["outputParameters"]["types"] = ["Trained Model"]
            node["data"]["color"] = "teal.600"
            graph["edges"].append(
                {
                    "source": node["id"],
                    "target": prediction_id,
                    "sourceHandle": "output_0_Trained Model",
                    "targetHandle": "input_1_Trained Model",
                    "type": "custom",
                    "id": f"reactflow_edge-{node['id']}output_0_Trained Model-{prediction_id}input_1_Trained Model",
                    "data": {
                        "sourceColor": "teal.600",
                        "targetColor": "green.400",
                        "target": prediction_id + "_input_1_Trained Model",
                        "source": node["id"] + "_output_0_Trained Model",
                    },
                }
            )

    graph["edges"] = [edge for edge in graph["edges"] if edge["target"] not in ["998", id]]
    graph["nodes"] = [
        node
        for node in graph["nodes"]
        if node["data"]["node_type"]
        in [
            "primary_dataset",
            "preprocess",
            "essential",
            "classification_model",
            "clusterring_model",
            "regression_model",
        ]
    ]

    return graph



import pandas as pd
from dateutil.parser import parse

def get_dataframe_types(df):
    data_dict = pd.io.json.build_table_schema(df)  # type: ignore
    cat_cols = [col for col in df.columns if df[col].nunique() < 10]
    date_formats = ['%Y-%m-%d', '%Y/%m/%d', '%m-%d-%Y', '%m/%d/%Y', '%Y-%m-%d %H:%M:%S', '%Y/%m/%d %H:%M:%S', '%m-%d-%Y %H:%M:%S', '%m/%d/%Y %H:%M:%S']

    date_cols = []

    for col in df.columns:
        if df.dtypes[col]=='O':
          try:
              if df[col].notnull().any() and any([isinstance(parse(str(val), fuzzy=False), (pd.Timestamp, pd.datetime)) for val in df[col] if not pd.isna(val)] for fmt in date_formats):
                date_cols.append(col)
          except ValueError:
            pass
    for col in date_cols:
      df[col]=pd.to_datetime(df[col], errors = 'coerce')   
    for field in data_dict["fields"]:
        if field["type"] == "integer":
            field["type"] = "number"
        if field["name"] in cat_cols:
            field["type"]="categorical"
    data_dict["columns"] = data_dict.pop("fields")
    data_dict.pop("primaryKey")
    data_dict.pop("pandas_version")
    data_dict['columns'].pop(0)
    return data_dict



import pandas as pd
from sklearn.feature_selection import SelectFromModel
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import LabelEncoder
from sklearn.ensemble import RandomForestRegressor
from evalml.problem_types import detect_problem_type

def select_features(data: pd.DataFrame, target_column: str) -> pd.DataFrame:
    
    label_encoder = LabelEncoder()
    for col in data.select_dtypes(include='object'):
        data[col] = label_encoder.fit_transform(data[col].astype(str))

    X = data.drop(target_column, axis=1)
    y = data[target_column]
    task_type = str(detect_problem_type(y))
    if task_type=="regression":
      estimator = RandomForestRegressor(n_estimators=100, random_state=42)
    else:
      estimator = RandomForestClassifier(n_estimators=100, random_state=42)
    sfm = SelectFromModel(estimator)

    sfm.fit(X, y)
    selected_features_mask = sfm.get_support()
    selected_features_names = X.columns[selected_features_mask]

    selected_data = data[selected_features_names.union([target_column])]

    return selected_data, selected_features_names.to_list()



import numpy as np
import pandas as pd
from scipy import stats
import evalml
from evalml.problem_types import detect_problem_type
from evalml.data_checks import (
    DataChecks,
    NullDataCheck,
    InvalidTargetDataCheck,
    NoVarianceDataCheck,
    ClassImbalanceDataCheck,
    TargetLeakageDataCheck,
    NoVarianceDataCheck,
    IDColumnsDataCheck,
    OutliersDataCheck,
    MulticollinearityDataCheck
)
from evalml.problem_types import ProblemTypes

def verify_ml_dataset(dataset, target):
    y_task = pd.Series(dataset[target])
    task=(detect_problem_type(y_task))
    task_str = str(task)
    if task_str=="regression":
        data_check_params={
            "InvalidTargetDataCheck": {
                "problem_type": ProblemTypes.REGRESSION,
                "objective": "R2",
            }
        }
    elif task_str=="binary":
        data_check_params={
            "InvalidTargetDataCheck": {
                "problem_type": ProblemTypes.BINARY,
                "objective": "log loss binary",
            }
        }

    elif task_str=="multiclass":
        data_check_params={
            "InvalidTargetDataCheck": {
                "problem_type": ProblemTypes.MULTICLASS,
                "objective": "log loss multiclass",
            }
        }
    else:
        data_check_params={}

    custom_data_checks = DataChecks(
        data_checks=[
            NullDataCheck,
            InvalidTargetDataCheck,
            NoVarianceDataCheck,
            ClassImbalanceDataCheck,
            TargetLeakageDataCheck,
            NoVarianceDataCheck,
            IDColumnsDataCheck,
            OutliersDataCheck,
            MulticollinearityDataCheck
        ],

        data_check_params=data_check_params
    )

    X = dataset.drop([target], axis=1)
    y = dataset[target]
    data_checks = custom_data_checks
    messages = data_checks.validate(X, y)
    return messages























    # # check for null values in pandas dataframe
    # DATA_IMBALANCE_THRESHOLD = 0.1
    # TOO_MANY_ZEROES_THRESHOLD = 0.1
    # FEATURE_IMPORTANCE_THRESHOLD = 0.01
    # dict_main = {}
    # dict_main["errors"] = {"null_values": [], "string_features": []}
    # dict_main["warnings"] = {
    #     "too_many_zeros": [],
    #     "outliers": [],
    #     "data_imbalance": [],
    #     "data_normalization": [],
    # }
    # dict_main["success"] = {
    #     "null_values": [],
    #     "string_features": [],
    #     "too_many_zeros": [],
    #     "outliers": [],
    #     "data_imbalance": [],
    #     "data_normalization": [],
    # }
    # dict_main["status"] = "ok"

    # for column in dataset.columns:

    #     # check for null values in pandas dataframe
    #     if dataset[column].isnull().values.any():
    #         dict_main["errors"]["null_values"].append(column)
    #         dict_main["status"] = "error"
    #     else:
    #         dict_main["success"]["null_values"].append(column)

    #     # check for string columns in pandas dataframe
    #     if (dataset[column].dtypes == object) == False:
    #         dict_main["errors"]["string_features"].append(column)
    #         dict_main["status"] = "error"
    #     else:
    #         dict_main["success"]["string_features"].append(column)

    #     # check for too many zeroes in pandas dataframe
    #     count = 0
    #     column_name = dataset[column]
    #     count += (column_name == 0).sum()
    #     if count > len(dataset[column]) * TOO_MANY_ZEROES_THRESHOLD:
    #         dict_main["warnings"]["too_many_zeros"].append(column)
    #         break

    #     # Check for outliers
    #     if column in dict_main["success"]["string_features"]:
    #         z_scores = stats.zscore(dataset[column])
    #         abs_z_scores = np.abs(z_scores)
    #         # Select data points with a z-scores above or below 3
    #         # filtered_entries = (abs_z_scores < 3)
    #         # print(filtered_entries)
    #         if (abs_z_scores < 3).all():
    #             dict_main["warnings"]["outliers"].append(column)
    #         else:
    #             dict_main["success"]["outliers"].append(column)

    #     # Data normalization
    #     if column in dict_main["success"]["string_features"]:
    #         if dataset[column].min() < -1 or dataset[column].max() > 1:
    #             dict_main["warnings"]["data_normalization"].append(column)
    #             break
    #         else:
    #             dict_main["success"]["data_normalization"].append(column)

    # values = dataset[target].value_counts()
    # if (
    #     values[0] * DATA_IMBALANCE_THRESHOLD > values[1]
    #     or values[1] * DATA_IMBALANCE_THRESHOLD > values[0]
    # ):
    #     dict_main["warnings"]["data_imbalance"].append("data_imbalance")
    # else:
    #     dict_main["success"]["data_imbalance"].append("data_imbalance")

    # return dict_main.popitem()




class PreprocessHandler:
    def __init__(self) -> None:
        pass

    def return_function(self, function_name: str, params: dict):
        operations = {
            "Remove NAN values": self.__remove_nan_values,
            "Replace NAN values": self.__replace_nan_values,
            "Dropping rows or columns": self.__drop_rows_columns,
            "Remove Outliers": self.__remove_outliers,
            "Drop Duplicates": self.__drop_duplicates,
            "Change Data Type": self.__change_data_type,
            "Round Data": self.__round_data,
            "Filter DataFrame": self.__filter_dataframe,
            "Truncate DataFrame": self.__truncate_dataframe,
            "Sort Values": self.__sort_values,
            "Transpose DataFrame": self.__transpose,
            "Min Max Scaler": self.__min_max_scale,
            "Max Abs Scaler": self.__max_abs_scale,
            "Robust Scaler": self.__robust_scale,
            "Standard Scaler": self.__standard_scale,
            "Normalization": self.__normalize,
            "Ordinal Encoding": self.__ordinal_encode,
            "One Hot Encoding": self.__one_hot_encode,
            "Remove Mismatch Data": self.__remove_mismatch_data,
            "Rename Columns": self.__rename_columns,
            "Select Columns": self.__select_columns,
            "Clean Column Names": self.__clean_column_names,
            "Clip": self.__clip,
            "Merge Datasets": self.__merge,
            "Clean Dataset": self.__clean_data,
            "Split Datasets": self.__split,
        }
        return operations[function_name](params)

    def __remove_nan_values(self, params):
        subset = []
        for i in range(len(params["subset"])):
            subset.append(params["subset"][i]["value"])
        parameters = {
            "axis": params["axis"],
            "how": params["how"],
            "inplace": params["inplace"],
            "subset": subset,
        }
        return (
            f"data = huble.sklearn.remove_nan_values(data=data,parameters={parameters})"
        )

    def __replace_nan_values(self, params):
        parameters = {
            "missing_values": params["missing_values"],
            "strategy": params["strategy"],
            "fill_value": params["fill_value"],
        }
        return f"data = huble.sklearn.replace_nan_values(data=data, column='{params['column']}', parameters={parameters})"

    def __drop_rows_columns(self, params):
        labels = []
        for i in range(len(params["labels"])):
            labels.append(params["labels"][i]["value"])

        parameters = {
            "labels": labels,
            "axis": params["axis"],
            "inplace": params["inplace"],
            "errors": params["errors"],
        }
        return (
            f"data = huble.sklearn.drop_rows_columns(data=data,parameters={parameters})"
        )

    def __remove_outliers(self, params):
        columns = []
        for i in range(len(params["columns"])):
            columns.append(params["columns"][i]["value"])
        return f"data = huble.sklearn.remove_outliers(data=data,columns={columns})"

    def __drop_duplicates(self, params):
        subset = []
        if len(params["subset"]) != 0:
            for i in range(len(params["subset"])):
                subset.append(params["subset"][i]["value"])
        parameters = {
            "subset": subset,
            "keep": params["keep"],
            "inplace": params["inplace"],
            "ignore_index": params["ignore_index"],
        }
        return (
            f"data = huble.sklearn.drop_duplicates(data=data,parameters={parameters})"
        )

    def __change_data_type(self, params):
        parameters = {
            "dtype": params["data type"],
            "copy": params["copy"],
            "errors": params["errors"],
        }
        return f"data = huble.sklearn.change_data_type(data=data,column={params['column']}, parameters={parameters})"

    def __round_data(self, params):
        parameters = {
            "decimals": params["decimals"],
        }
        return f"data = huble.sklearn.round_data(data=data, parameters={parameters})"

    def __filter_dataframe(self, params):
        items = []
        for i in range(len(params["items"])):
            items.append(params["items"][i]["value"])
        parameters = {
            "items": items,
            "like": params["like"],
            "axis": params["axis"],
        }
        return (
            f"data = huble.sklearn.filter_dataframe(data=data, parameters={parameters})"
        )

    def __truncate_dataframe(self, params):
        parameters = {
            "before": params["before"],
            "after": params["after"],
            "copy": params["copy"],
            "axis": params["axis"],
        }
        return f"data = huble.sklearn.truncate_datfarame(data=data, parameters={parameters})"

    def __sort_values(self, params):
        by = []
        for i in range(len(params["by"])):
            by.append(params["by"][i]["value"])
        parameters = {
            "by": by,
            "axis": params["axis"],
            "ascending": params["ascending"],
            "inplace": params["inplace"],
            "kind": params["kind"],
            "na_position": params["na_position"],
            "ignore_index": params["ignore_index"],
        }
        return f"data = huble.sklearn.sort_values(data=data, parameters={parameters})"

    def __transpose(self):
        return f"data = huble.sklearn.transpose(data=data)"

    def __min_max_scale(self, params):
        columns = []
        for i in range(len(params["column"])):
            columns.append(params["column"][i]["value"])
        parameters = {
            "feature_range": params["feature_range"],
            "copy": params["copy"],
            "clip": params["clip"],
        }
        return f"data = huble.sklearn.min_max_scalar(data=data, columns={columns}, parameters={parameters})"

    def __max_abs_scale(self, params):
        columns = []
        for i in range(len(params["column"])):
            columns.append(params["column"][i]["value"])
        parameters = {
            "copy": params["copy"],
        }
        return f"data = huble.sklearn.max_abs_scalar(data=data, columns={columns}, parameters={parameters})"

    def __robust_scale(self, params):
        columns = []
        for i in range(len(params["column"])):
            columns.append(params["column"][i]["value"])
        parameters = {
            "with_centering": params["with_centering"],
            "with_scaling": params["with_scaling"],
            "copy": params["copy"],
            "unit_variance": params["unit_variance"],
            "quantile_range": params["quantile_range"],
        }
        return f"data = huble.sklearn.robust_scalar(data=data, columns={columns}, parameters={parameters})"

    def __standard_scale(self, params):
        columns = []
        for i in range(len(params["column"])):
            columns.append(params["column"][i]["value"])
        parameters = {
            "copy": params["copy"],
            "with_mean": params["with_mean"],
            "with_std": params["with_std"],
        }
        return f"data = huble.sklearn.standard_scalar(data=data, columns={columns}, parameters={parameters})"

    def __normalize(self, params):
        columns=[]
        for i in range(len(params['column'])):
            columns.append(params['column'][i]['value'])
        
        parameters = {
            "norm": params["norm"],
            "copy": params["copy"],
        }
        return f"data = huble.sklearn.normalize(data=data, columns={columns}, parameters={parameters})"

    def __ordinal_encode(self, params):
        columns = []
        for i in range(len(params["columns"])):
            columns.append(params["columns"][i]["value"])
        parameters = {
            "categories": params["categories"],
            "dtype": params["dtype"],
            "handle_unknown": params["handle_unknown"],
            "unknown_value": params["unknown_value"],
            "encoded_missing_value": params["encoded_missing_value"],
        }
        return f"data = huble.sklearn.ordinal_encode(data=data, columns={columns}, parameters={parameters})"

    def __one_hot_encode(self, params):
        columns = []
        for i in range(len(params["columns"])):
            columns.append(params["columns"][i]["value"])
        parameters = {
            "categories": params["categories"],
            "dtype": params["dtype"],
            "handle_unknown": params["handle_unknown"],
            "sparse": params["sparse"],
            "min_frequency": params["min_frequency"],
            "max_categories": params["max_categories"],
        }
        return f"data = huble.sklearn.one_hot_encode(data=data, columns={columns}, parameters={parameters})"

    def __remove_mismatch_data(self, params):
        columns = []
        for i in range(len(params["exceptions"])):
            columns.append(params["exceptions"][i]["value"])
        parameters = {
            "exceptions": columns,
        }
        return f"data = huble.sklearn.remove_mismatch_data(data=data, parameters={parameters})"

    def __rename_columns(self, params):
        parameters = {
            "mapper": params["mapper"],
            "axis": params["axis"],
            "errors": params["errors"],
        }
        return (
            f"data = huble.sklearn.rename_columns(data=data, parameters={parameters})"
        )

    def __select_columns(self, params):
        include = []
        for i in range(len(params["include"])):
            include.append(params["include"][i]["value"])
        exclude = []
        for i in range(len(params["exclude"])):
            exclude.append(params["exclude"][i]["value"])
        parameters = {
            "include": include,
            "exclude": exclude,
        }
        return (
            f"data = huble.sklearn.select_columns(data=data, parameters={parameters})"
        )

    def __clean_column_names(self, params):
        return f"data = huble.sklearn.clean_column_names(data=data)"

    def __clip(self, params):
        parameters = {
            "lower": params["lower"],
            "upper": params["upper"],
            "axis": params["axis"],
        }
        return f"data = huble.sklearn.clip(data=data, parameters={parameters})"

    def __merge(self, params):
        on = []
        for i in range(len(params["on"])):
            on.append(params["on"][i]["value"])
        left_on = []
        for i in range(len(params["left_on"])):
            left_on.append(params["left_on"][i]["value"])
        right_on = []
        for i in range(len(params["right_on"])):
            right_on.append(params["right_on"][i]["value"])
        parameters = {
            "right": params["right"],
            "how": params["how"],
            "on": on,
            "left_on": left_on,
            "right_on": right_on,
            "sort": params["sort"],
        }
        return f"data = huble.sklearn.merge(data=data, parameters={parameters})"

    def __clean_data(self, params):
        return f"data = huble.sklearn.clean_data(data=data)"

    def __split(self, params):
        return f"data = huble.sklearn.split(data=data)"



from sklearn.model_selection import train_test_split as sklearn_tts  # type: ignore
import pandas as pd
from ...error.decorators import function_error_handling
from ...connector import Dataset
import joblib
from ...util.data_types import get_dataframe_types
from ...util.feature_selection import select_features

# @function_error_handling("train_test_split")
def train_test_split(**params):
    data = params["data"]
    target = params["target_column"]
    model = params["model"]
    X = data.drop([target], axis=1)
    data_dict = get_dataframe_types(X)
    data, selected_features = select_features(data, target)
    train_dataset, test_dataset = sklearn_tts(data, test_size=params["parameters"]["test_size"], random_state=42)
    return train_dataset, test_dataset, data_dict, selected_features


# @function_error_handling("train_model")
def train_model(**params):
    task_type = params["task_type"]
    data = params["data"]
    model = params["model"]
    if task_type == "clustering":
        Model = model.fit(data)
    else:
        y = data[params["column"]]
        X = data.drop(params["column"], axis=1)

        Model = model.fit(X, y)
    filename = f'model-{params["id"]}.joblib'
    joblib.dump(Model, filename)
    return Model, filename



from sklearn import metrics
import evalml
import pandas as pd
from evalml.problem_types import detect_problem_type
import numpy as np

def regression_metrics(y_true, y_pred):
    explained_variance_score=metrics.explained_variance_score(y_true, y_pred)
    max_error = metrics.max_error(y_true, y_pred)
    mean_absolute_error = metrics.mean_absolute_error(y_true, y_pred)
    mean_squared_error = metrics.mean_squared_error(y_true, y_pred)
    mean_squared_log_error = metrics.mean_squared_log_error(y_true, y_pred)
    median_absolute_error = metrics.median_absolute_error(y_true, y_pred)
    r2_score = metrics.r2_score(y_true, y_pred)
    rmse = np.sqrt(mean_squared_error)
    rmsle = np.sqrt(mean_squared_log_error)
   
    
    # create dictionary of metrics
    metrics_dict = {
        "ExpVariance":explained_variance_score,
        "MaxError": max_error,
        "MedianAE": median_absolute_error,
        "MAE": mean_absolute_error,
        "MSE": mean_squared_error,
        "R2": r2_score,
        "Mean Squared Log Error": mean_squared_log_error,
        "Root Mean Squared Log Error": rmsle,
        "Root Mean Squared Error": rmse, 
    }
    return metrics_dict


def classification_metrics(y_true, y_pred):
    accuracy_score = metrics.accuracy_score(y_true, y_pred)
    balanced_accuracy_score = metrics.balanced_accuracy_score(y_true, y_pred)
    f1_score = metrics.f1_score(y_true, y_pred)
    f1_weighted = metrics.f1_score(y_true, y_pred, average="weighted")
    f1_micro = metrics.f1_score(y_true, y_pred, average="micro")
    f1_macro = metrics.f1_score(y_true, y_pred, average="macro")
    log_loss = metrics.log_loss(y_true, y_pred)
    matthews_corrcoef = metrics.matthews_corrcoef(y_true, y_pred)
    precision_score = metrics.precision_score(y_true, y_pred)
    precision_weighted = metrics.precision_score(y_true, y_pred, average="weighted")
    precision_micro = metrics.precision_score(y_true, y_pred, average="micro")
    precision_macro = metrics.precision_score(y_true, y_pred, average="macro")
    recall_score = metrics.recall_score(y_true, y_pred)
    recall_weighted = metrics.recall_score(y_true, y_pred, average="weighted")
    recall_micro = metrics.recall_score(y_true, y_pred, average="micro")
    recall_macro = metrics.recall_score(y_true, y_pred, average="macro")
    roc_auc_score = metrics.roc_auc_score(y_true, y_pred)
    roc_auc_weighted = metrics.roc_auc_score(y_true, y_pred, average="weighted")
    roc_auc_micro = metrics.roc_auc_score(y_true, y_pred, average="micro")
    roc_auc_macro = metrics.roc_auc_score(y_true, y_pred, average='macro')
    
    metrics_dict={}
    y_task = pd.Series(y_true)
    task=(detect_problem_type(y_task))
    task_str = str(task)
    if task_str == 'binary':
        metrics_dict = {
            "MCC Binary": matthews_corrcoef,
            "Log Loss Binary": log_loss,
            "AUC": roc_auc_score,
            "Precision": precision_score,
            "Recall": recall_score,
            "F1": f1_score,
            "Accuracy Binary": accuracy_score,
            "Balanced Accuracy Binary": balanced_accuracy_score,      
        }
    elif task == 'multiclass':
        metrics_dict = {
            "MCC Multiclass": matthews_corrcoef,
            "Log Loss Multiclass": log_loss,
            "Accuracy Multiclass": accuracy_score,
            "Balanced Accuracy Multiclass": balanced_accuracy_score,
            "Precision Weighted": precision_weighted,
            "Precision Micro": precision_micro,
            "Precision Macro": precision_macro,
            "Recall Weighted": recall_weighted,
            "Recall Micro": recall_micro,
            "Recall Macro": recall_macro,            
            "F1 Weighted": f1_weighted,   
            "F1 Macro": f1_macro, 
            "F1 Mirco": f1_micro,  
            "AUC Weighted": roc_auc_weighted,
            "AUC Micro": roc_auc_micro,
            "AUC Macro": roc_auc_macro, 
        }
    return metrics_dict


def clustering_metrics(X, labels):
    silhouette_score = metrics.silhouette_score(X, labels)
    calinski_harabasz_score = metrics.calinski_harabasz_score(X, labels)
    davies_bouldin_score = metrics.davies_bouldin_score(X, labels)
    
    # create dictionary of metrics
    metrics_dict = {
        "silhouette_score": silhouette_score,
        "calinski_harabasz_score": calinski_harabasz_score,
        "davies_bouldin_score": davies_bouldin_score,
    }
    return metrics_dict


def log_metrics(y_true, y_pred, task, X, labels):
    if task == "regression":
        metrics_dict = regression_metrics(y_true, y_pred)
        return metrics_dict
    elif task == "classification":
        metrics_dict = classification_metrics(y_true, y_pred)
        return metrics_dict
    elif task == "clustering":
        metrics_dict = clustering_metrics(X, labels)
        return metrics_dict



from .verify_ml import verify_ml_dataset
from .recommend_model import recommend_model
from .preprocess_code import preprocess_code_generation
from .convert_to_reactflow_graph import convert_to_graph as get_reatflow_graph
from .typed_params import get_typed_params
from .inference_pipeline import inference_pip
from .feature_selection import select_features



import requests
from boto3.session import Session
import json


class Experiment:
    def __init__(self, experiment_id: str) -> None:
        self.experiment_id = experiment_id
        (
            self.graph,
            self.task_type,
            self.target_column,
            self.modelURL,
            self.input_format,
            self.selected_features
        ) = self.__get_experiment_details(experiment_id=experiment_id)

    def __get_experiment_details(self, experiment_id: str):
        try:
            response = requests.get(
                f"http://localhost:8000/experiments/{experiment_id}",
            )
            response = response.json()
        except:
            try:
                response = requests.get(
                    f"http://const_server:8000/experiments/{experiment_id}",
                )
                response = response.json()
            except:
                raise Exception("Unable to connect to the server")
        try:
            graph = response["pipelineJSON"]
            project = response["project"]
            target_column = project["targetColumn"]
            task_type = project["taskType"]
            modelURL = response["modelURL"]
            input_format = response["inputFormat"]
            selected_features = response["selectedFeatures"]
        except:
            raise Exception("Invalid Experiment ID")
        return graph, task_type, target_column, modelURL, input_format,selected_features

    def upload_metrics(self, metrics, input_format, feature_importance_dict, selected_features):
        requests.put(
            f"http://localhost:8000/experiments/results/{self.experiment_id}",
            data={
                "metrics": json.dumps(metrics),
                "input_format": json.dumps(input_format),
                "feature_importance": json.dumps(feature_importance_dict),
                "selected_features": json.dumps(selected_features),
            },
        )

    def upload_model(self, file_name):
        try:
            with open(file_name, "rb") as data:
                print("Uploading model...")
                requests.put(
                    f"http://localhost:8000/experiments/{self.experiment_id}/model/upload", files={"file": data}
                )
            print("Model uploaded successfully")
        except:
            Exception("Unable to upload model")
